
listeners <- []
destroyed_listeners <- []


monster_spawn_enabled<-true


function reset_listeners()
{
  listeners.clear()
  destroyed_listeners.clear()
}


function destroy_listeners()
{
  foreach (l in destroyed_listeners)
    foreach (i,v in listeners)
      if (v==l) { listeners.remove(i); break; }

  destroyed_listeners.clear()
}


function find_listener(filter)
{
  foreach (l in listeners)
    if (!l.isDestroyed && filter(l)) return l;
  return null;
}


class Listener
{
  </persistent=1/> isDestroyed=false

  constructor(params=null)
  {
    if (params) foreach (n,v in params)
      if (n in this) this[n]=v

    listeners.append(this)
  }

  function destroy()
  {
    if (isDestroyed) return;
    isDestroyed=true
    destroyed_listeners.append(this)
  }

  function _processEvent(event)
  {
    if (isDestroyed) return;
    processEvent(event)
  }

  function processEvent(event) {}
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class SingleEventListener extends Listener
{
  </persistent=1/> event=null
  </persistent=1/> once=true

  constructor(e, params=null)
  {
    event=e
    base.constructor(params)
  }

  function processEvent(e)
  {
    if (e==event)
    {
      onEvent()
      if (once) destroy()
    }
  }

  function onEvent() {}
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class MarkInactiveListener extends SingleEventListener
{
  </persistent=1/> objects=null

  constructor(e, o, params=null)
  {
    objects=o
    base.constructor(e, params)
  }

  function onEvent()
  {
    if (objects) foreach (o in objects) o.markInactive()
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class RemoveRewardListener extends SingleEventListener
{
  </persistent=1/> rewardName=null
  </persistent=1/> rewardAmount=0

  constructor(e, n, a, params=null)
  {
    base.constructor(e, params)
    rewardName=n
    rewardAmount=a
  }

  function onEvent()
  {
    if (!cur_world || !rewardName) return;
    if (!(rewardName in cur_world.remainingRewards)) return;
    cur_world.remainingRewards[rewardName]-=rewardAmount
    if (cur_world.remainingRewards[rewardName]<=0)
      delete cur_world.remainingRewards[rewardName]
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class TriggerSoundListener extends SingleEventListener
{
  </persistent=1/> sound="trigger_activated"

  function onEvent()
  {
    play_sound_event(sound)
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class ShowMessageListener extends SingleEventListener
{
  </persistent=1/> text="Oh hey!"
  </persistent=1/> subtext=""

  function onEvent()
  {
    show_screen_message(text)
    show_screen_sub_message(subtext)
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class SingleSpawnListener extends SingleEventListener
{
  </persistent=1/> posX=0
  </persistent=1/> posZ=0

  topo=null
  ox=0
  oz=0

  constructor(e, params=null)
  {
    base.constructor(e, params)

    if (topo!=null)
    {
      posX=topo.ox+ox
      posZ=topo.oz+oz
      topo=null
    }
  }

  function onEvent()
  {
    spawnObject(posX, posZ)
  }

  function spawnObject(x, z) {}
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class SpawnListener extends SingleEventListener
{
  </persistent=1/> number=10
  </persistent=1/> mask=null

  topo=null
  maskName="mask"

  constructor(e, params=null)
  {
    base.constructor(e, params)

    if (topo!=null)
    {
      mask=topo[maskName].shiftedCopy(topo.ox, topo.oz)
      topo=null
    }
  }

  function onEvent()
  {
    if (mask!=null) mask.placePoints(number, spawnObject.bindenv(this))
  }

  function spawnObject(x, z) {}
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


function spawn_loot_mana(amount, points)
{
  if (points.len()==2)
  {
    ManaBall(points[0], amount, points[1]*0.5+vec4f(srnd()*2, (frnd()+1)*3, srnd()*2, 0))
  }
  else
  {
    for (local i=0; i<points.len(); i+=2)
    {
      local a
      if (i==0) a=amount*0.5
      else a=amount*0.5/(points.len()/2-1)
      ManaBall(points[i], a, points[i+1]*0.5+vec4f(srnd()*2, (frnd()+1)*3, srnd()*2, 0))
    }
  }
}


class MonsterSpawnListener extends SpawnListener
{
  </persistent=1/> lootFunc=null
  </persistent=1/> lootData=0.1

  constructor(e, params=null)
  {
    base.constructor(e, params)
    if (lootFunc==null) lootFunc=spawn_loot_mana
  }
}


class SpawnBirdsListener extends MonsterSpawnListener
{
  </persistent=1/> number=40
  </persistent=1/> lootData=0.05

  function spawnObject(x, z)
  {
    if (!monster_spawn_enabled) return;
    local m=BirdMonster(vec4f(x, get_land_height(x, z)-5, z, 1), make_dir_xz(frnd()))
    m.setLoot(lootFunc, lootData)
  }
}


class SpawnDarkBirdsListener extends MonsterSpawnListener
{
  </persistent=1/> number=40
  </persistent=1/> lootData=0.05

  function spawnObject(x, z)
  {
    if (!monster_spawn_enabled) return;
    local m=BirdMonster(vec4f(x, get_land_height(x, z)-5, z, 1), make_dir_xz(frnd()))
    m.setBodyModel("bird", 0xffff1100)
    m.setLoot(lootFunc, lootData)
  }
}


class SpawnCaterpillarsListener extends MonsterSpawnListener
{
  </persistent=1/> number=5
  </persistent=1/> lootData=0.5

  function spawnObject(x, z)
  {
    if (!monster_spawn_enabled) return;
    local m=LandMonster(vec4f(x, get_land_height(x, z)-5, z, 1), make_dir_xz(frnd()))
    m.setLoot(lootFunc, lootData)
  }
}


class SpawnWyvernsListener extends MonsterSpawnListener
{
  </persistent=1/> number=3
  </persistent=1/> lootData=1

  function spawnObject(x, z)
  {
    if (!monster_spawn_enabled) return;
    local m=WyvernMonster(vec4f(x, get_land_height(x, z)-10, z, 1))
    m.setLoot(lootFunc, lootData)
  }
}


function get_monster_spawn_listener()
{
  local r=frnd()
  if (r<0.33) return SpawnWyvernsListener;
  else if (r<0.66) return SpawnCaterpillarsListener;
  else if (wgen && ("dark" in wgen.tags)) return SpawnDarkBirdsListener;
  else return SpawnBirdsListener;
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class SpawnManaListener extends SpawnListener
{
  </persistent=1/> amount=2
  </persistent=1/> randomness=0.5
  </persistent=1/> ownerId=0

  function spawnObject(x, z)
  {
    local a=1+srnd()*randomness
    if (a<0.01) return;

    local b=ManaBall(
      vec4f(x, get_land_height(x, z)-10, z, 1),
      a*amount/number,
      vec4f(0, rndf(15, 30), 0, 0))
    b.setOwnerId(ownerId)
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class SpawnSpellListener extends SingleSpawnListener
{
  </persistent=1/> spellName="FireballSpell"

  constructor(e, s, params=null)
  {
    spellName=s
    base.constructor(e, params)
  }

  function spawnObject(x, z)
  {
    SpellObject(x, z, spellName)
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class SpawnGlobResListener extends SingleSpawnListener
{
  </persistent=1/> resName=null
  </persistent=1/> amount=0

  constructor(e, n, a, params=null)
  {
    resName=n
    amount=a
    base.constructor(e, params)
  }

  function spawnObject(x, z)
  {
    GlobResObject(x, z, resName, amount)
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class SpawnPortalListener extends SingleSpawnListener
{
  function spawnObject(x, z)
  {
    WorldPortal(vec4f(x, get_land_height(x, z), z, 1))
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class SpawnMenhirListener extends SingleSpawnListener
{
  function spawnObject(x, z)
  {
    (frnd()<0.5?Menhir1StaticObject:Menhir2StaticObject)(x, z).makeSpecial()
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class StoneCupCircleTrigger extends Listener
{
  </persistent=1/> objects=null
  </persistent=1/> event=null
  </persistent=1/> once=true

  constructor(o, e, params=null)
  {
    objects=o
    event=e
    base.constructor(params)
  }

  function processEvent(e)
  {
    if (e=="update")
    {
      local ok=true
      foreach (o in objects)
        if (!o.isBurning()) { ok=false; break; }

      if (ok)
      {
        foreach (o in objects) o.quenchFire()
        process_event(event)
        if (once)
        {
          foreach (o in objects) { o.setActive(false); o.markInactive() }
          destroy()
        }
      }
    }
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class TreeCircleTrigger extends Listener
{
  </persistent=1/> objects=null
  </persistent=1/> event=null
  </persistent=1/> once=true

  constructor(o, e, params=null)
  {
    objects=o
    event=e
    base.constructor(params)
  }

  function processEvent(e)
  {
    if (e=="update")
    {
      local ok=true
      foreach (o in objects)
        if (!o.hasFoliage()) { ok=false; break; }

      if (ok)
      {
        process_event(event)
        if (!once)
          foreach (o in objects) o.setFoliage(false)
        else
        {
          foreach (o in objects) o.markInactive()
          destroy()
        }
      }
    }
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class TriggerHub extends Listener
{
  </persistent=1/> completionEvent=null
  </persistent=1/> events=null
  </persistent=1/> maxMonsters=0

  function processEvent(e)
  {
    if (e=="update")
    {
      if (!events || events.len()==0)
        if (count_alive_monsters()<=maxMonsters)
        {
          process_event(completionEvent)
          destroy()
        }
    }
    else if (events)
    {
      for (local i=events.len()-1; i>=0; --i)
        if (e==events[i]) events.remove(i)
    }
  }
}
