

class EndOfDemoWorldGen extends GenericWorldGen
{
  tags="WorldGen end_of_demo normal"

  function generate()
  {
    WELEM("!PlayerStart")().addToWGen(this, "top")

    local trigger=SingleMenhir({onEvent="message"})
    trigger.addToWGen(this, "top")
    removeElem(trigger.topo, "topTopo")
    addElem(TopoIsland(trigger.topo, {extraSteps=300}), "topTopo")

    WELEM("!WorldShaper islands")().shapeWorld(this)

    ShowMessageListener("message", {once=false,
      text="Hmm. It looks like the end of the DEMO universe!"})
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class PesterWithMessage extends Listener
{
  </persistent=1/> text="Oh hey! Do something!"
  </persistent=1/> subtext=""
  </persistent=1/> timer=10
  </persistent=1/> period=30

  function processEvent(e)
  {
    if (e=="update")
    {
      if (--timer<=0)
      {
        timer=period
        show_screen_message(text)
        show_screen_sub_message(subtext)
      }
    }
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class TutorialController extends Listener
{
  </persistent=1/> pester=null
  </persistent=1/> altar=null
  </persistent=1/> firebloom=null
  </persistent=1/> arrow=null
  </persistent=1/> altarArrow=null
  </persistent=1/> stage=""

  function processEvent(e)
  {
    if (e=="update")
    {
      switch (stage)
      {
        case "mana1":
          if (player_has_spell_selected("ClaimManaSpell"))
          {
            stage="mana2"
            pester.text="I need to claim that wild mana to power my spells."
            pester.subtext=format(
              "(Use %s or %s to cast\nthe selected spell. Hold %s to show the map.)",
              get_assigned_button_name(B_lspell),
              get_assigned_button_name(B_rspell),
              get_assigned_button_name(B_map))
            pester.timer=7

            if (arrow) { arrow.destroy(); arrow=null }
          }
          break;

        case "mana2":
          local mball=null
          for_each_game_object(function(o)
          {
            if (o.hasAspect(ID_ManaBall) && o.getOwnerId()==0) { mball=o; return false; }
            return true;
          })

          if (mball)
          {
            if (!arrow) arrow=MapArrow(vec4f(0, 0, 0, 1), 0xcccc1111)
            arrow.setPos(mball.getPos())
          }
          else
          {
            if (arrow) { arrow.destroy(); arrow=null }
          }

          if (get_owned_mana(1)>=2.0)
          {
            stage="firebloom"
            pester.text="I need more spells. There's a useful ingredient here."
            pester.subtext=format("(Hold %s to show the map.)", get_assigned_button_name(B_map))
            pester.timer=7

            if (arrow) { arrow.destroy(); arrow=null }
            if (firebloom) arrow=MapArrow(vec4f(firebloom.posX, 0, firebloom.posZ, 1), 0xcccc1111)

            process_event("mana_claimed")
          }
          break;

        case "firebloom":
          if (get_glob_res_amount("firebloom")>0)
          {
            stage="fireball"
            pester.text="Excellent! Now I can make a new spell at the altar."
            pester.subtext=format("(Hold %s to show the map.)", get_assigned_button_name(B_map))
            pester.timer=7

            if (arrow) { arrow.destroy(); arrow=null }
            altarArrow=MapArrow(vec4f(altar.posX, 0, altar.posZ, 1), 0xcc11cc11)
          }
          break;

        case "fireball":
          if (player_has_spell("FireballSpell"))
          {
            get_local_player().setSpell(0, ID_FireballSpell)
            stage="monster1"
            pester.text="A monster is seeking to destroy my castle. I should dispose of it."
            pester.subtext=format("(Use %s to cast left spell.)", get_assigned_button_name(B_lspell))
            pester.timer=7
            process_event("spawn1")
          }
          break;

        case "monster1":
          if (!are_any_monsters_alive())
          {
            stage="monster2"
            pester.text="Oh no! More monsters! At least they drop mana."
            pester.subtext=""
            pester.timer=7
            process_event("spawn2")
          }
          break;

        case "monster2":
          if (!are_any_monsters_alive())
          {
            stage="portal1"
            pester.text="Oh! A portal! But I need a spell to use it..."
            pester.subtext=format("(Hold %s to show the map.)", get_assigned_button_name(B_map))
            pester.timer=7

            process_event("spawn_portal")

            altar.recipeType="spell"
            altar.recipeName="UsePortalSpell"
            altar.recipeCosts={ vortex=1 }
          }
          break;

        case "portal1":
          if (player_has_spell("UsePortalSpell"))
          {
            stage="portal2"
            pester.text="Great. Now I can use the portal to explore more worlds."
            pester.subtext=format("(Hold %s to select spells.)", get_assigned_button_name(B_spells))
            pester.timer=7

            if (arrow) { arrow.destroy(); arrow=null }
            if (altarArrow) { altarArrow.destroy(); altarArrow=null }
          }
          break;
      }
    }
    else if (e=="got_castle_spell")
    {
      get_local_player().setSpell(1, ID_CastleSpell)
      pester.text="Good. Now I need to establish my base here."
      pester.subtext=format("(Use %s to cast right spell.)", get_assigned_button_name(B_rspell))
      pester.timer=7

      if (arrow) { arrow.destroy(); arrow=null }
    }
    else if (e=="castle_placed")
    {
      stage="mana1"
      pester.text="Ah, there's some wild mana here. I have a spell to claim it."
      pester.subtext=format("(Hold %s to select spells, then click Claim Mana spell.)", get_assigned_button_name(B_spells))
      pester.timer=7
    }
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class TutorialWorldGen extends GenericWorldGen
{
  tags="WorldGen starting_world normal"

  function generate()
  {
    WELEM("!PlayerStart")().addToWGen(this, "top")

    local spellPlace=TopoCircle(5)
    addElem(spellPlace, "topTopo")

    local manaArea=WELEM("!Area")()
    manaArea.addToWGen(this, "top")
    BonusMana({event="got_castle_spell", targetTopo=manaArea.topo,
      amount=5, number=30, ownerId=0}).addToWGen(this, "top")

    local firebloomPlace=StoneCircle()
    firebloomPlace.addToWGen(this, "top")
    local firebloomReward=ResourceReward({firebloom=1}, "mana_claimed",
      {targetTopo=firebloomPlace.topo})
    firebloomReward.addToWGen(this, "top")

    local vortexPlace=StoneCircle()
    vortexPlace.addToWGen(this, "top")
    ResourceReward({vortex=1}, "spawn_portal",
      {targetTopo=vortexPlace.topo}).addToWGen(this, "top")

    WELEM("!Reward !r_portal")({portal=1}, "spawn_portal").addToWGen(this, "top")

    local altar=WELEM("!CraftingAltar")(0)
    altar.addToWGen(this, "top")
    removeElem(altar.topo, "topTopo")

    local monsterTopo=TopoIsland(altar.topo, {extraSteps=10})
    addElem(monsterTopo, "topTopo")

    local monsterArea=WELEM("!Area")()
    monsterArea.addToWGen(this, "top")

    WELEM("!WorldShaper islands")().shapeWorld(this)

    altar.listener.recipeType="spell"
    altar.listener.recipeName="FireballSpell"
    altar.listener.recipeCosts={ firebloom=1 }

    // MapArrow(vec4f(altar.geo.posX(), 0, altar.geo.posZ(), 1), 0xcc11cc11)

    local o=SpellObject(spellPlace.ox, spellPlace.oz, "CastleSpell")
    o.setEvent("got_castle_spell")

    local arrow=MapArrow(o.getPos(), 0xcccc1111)

    SpawnCaterpillarsListener("spawn1", { topo=monsterTopo, number=1 })
    SpawnBirdsListener("spawn2", { topo=monsterArea.topo, number=15 })

    TriggerSoundListener("spawn1")
    TriggerSoundListener("spawn2")

    TutorialController({
      pester=PesterWithMessage({text=
        "There's a spell in this world, I must find it.",
        subtext=format(
          "(Use %s, %s, %s and %s to move. Double-tap or hold %s to dash.)",
          get_assigned_button_name(B_forward),
          get_assigned_button_name(B_back),
          get_assigned_button_name(B_left),
          get_assigned_button_name(B_right),
          get_assigned_button_name(B_turbo)
        )})
      altar=altar.listener
      firebloom=firebloomReward.listener
      arrow=arrow
    })
  }
}
